from dataclasses import dataclass
import math

from .cmyk import CMYKMixin
from .rgb import RGBMixin
from .hsv import HSVMixin
from .hsl import HSLMixin
from .hex import HexMixin
from .lab import LabMixin
from .xyz import XYZMixin


@dataclass
class Colour(CMYKMixin, RGBMixin, HSLMixin, HSVMixin, LabMixin, XYZMixin, HexMixin):
    """
    A class to easily convert between many kinds of colour models, including:

    - RGB, RGBA, hex
    - HSL
    - HSV
    - CMYK
    - CIELAB
    - CIEXYZ

    Internally it stores RGBA colour values as an intermediate between colour models.
    These values are accessable from properties .r, .g, .b and .a.

    Made by sertdfyguhi (https://github.com/sertdfyguhi).
    """

    r: int | float
    g: int | float
    b: int | float
    a: int | float = 255

    def floor(self) -> None:
        """
        Floors the internal RGBA colour values.

        Returns: None
        """
        self.r = math.floor(self.r)
        self.g = math.floor(self.g)
        self.b = math.floor(self.b)
        self.a = math.floor(self.a)

    def round(self) -> None:
        """
        Rounds the internal RGBA colour values.

        Returns: None
        """
        self.r = round(self.r)
        self.g = round(self.g)
        self.b = round(self.b)
        self.a = round(self.a)

    def ceil(self) -> None:
        """
        Ceils the internal RGBA colour values.

        Returns: None
        """
        self.r = math.ceil(self.r)
        self.g = math.ceil(self.g)
        self.b = math.ceil(self.b)
        self.a = math.ceil(self.a)



# standard d65
xn = 95.0489
yn = 100
zn = 108.8840


def normalize_XYZ(x: int | float, y: int | float, z: int | float) -> tuple[int | float]:
    """
    Normalizes XYZ colour values using CIE standard illuminant D65 as white achromatic reference illuminant.

    Args:
        x: int | float
            X value.
        y: int | float
            Y value.
        z: int | float
            Z value.

    Returns: tuple[int | float] consisting of normalized XYZ colour values.
    """
    return (x / xn, y / yn, z / zn)


class XYZMixin:
    """Mixin to add CIEXYZ conversion functions."""

    def to_xyz(self, normalize: bool = True) -> tuple[int | float]:
        """
        Converts the Colour object into XYZ. Assumes Colour object is in the sRGB colour space.
        Uses CIE standard illuminant D65 as white reference.

        Args:
            normalize: bool = True
                If true, normalizes XYZ colour values before returning.

        Returns: tuple[int | float] consisting of XYZ values.
        """
        # linearize sRGB data
        rgb = self.to_rgbd()
        rt, gt, bt = (
            c / 12.92 if c <= 0.04045 else ((c + 0.055) / 1.055) ** 2.4 for c in rgb
        )

        # transformation matrix
        x = 0.4124564 * rt + 0.3575761 * gt + 0.1804375 * bt
        y = 0.2126729 * rt + 0.7151522 * gt + 0.0721750 * bt
        z = 0.0193339 * rt + 0.1191920 * gt + 0.9503041 * bt

        if normalize:
            return (x, y, z)
        else:
            return (x * xn, y * yn, z * zn)

    @classmethod
    def from_xyz(cls, x: int | float, y: int | float, z: int | float):
        """
        Creates a Colour object from a XYZ colour.
        Uses CIE standard illuminant D65 as white reference.

        Args:
            x: int | float (0.0 - 1.0)
                Normalized X value.
            y: int | float (0.0 - 1.0)
                Normalized Y value.
            z: int | float (0.0 - 1.0)
                Normalized Z value.

        Returns: Colour
        """
        # transformation matrix
        rgbt = (
            3.2404542 * x + -1.5371385 * y + -0.4985314 * z,
            -0.9692660 * x + 1.8760108 * y + 0.0415560 * z,
            0.0556434 * x + -0.2040259 * y + 1.0572252 * z,
        )

        # apply gamma
        r, g, b = (
            c * 12.92 if c <= 0.0031308 else 1.055 * c ** (1 / 2.4) - 0.055
            for c in rgbt
        )

        return cls(r * 255, g * 255, b * 255)



import math

# constants
d = 6 / 29
d1 = d**3
d2 = 2 * d**2

p = 4 / 29


def f(t):
    return math.cbrt(t) if t > d1 else t / d2 + p


def inv_f(t):
    return t**3 if t > d else d2 * (t - p)


class LabMixin:
    """Mixin to add CIELAB conversion functions."""

    def to_lab(self) -> tuple[int | float]:
        """
        Converts the Colour object into Lab.

        Returns: tuple[int | float] consisting of Lab colour values.
        """
        x, y, z = self.to_xyz(normalize=True)
        yt = f(y)

        L = 116 * yt - 16
        a = 500 * (f(x) - yt)
        b = 200 * (yt - f(z))

        return (L, a, b)

    @classmethod
    def from_lab(cls, l: int | float, a: int | float, b: int | float):
        """
        Creates a Colour object from a Lab colour.

        Args:
            l: int | float
                L value.
            a: int | float
                a value.
            b: int | float
                b value.

        Returns: Colour
        """
        lp = (l + 16) / 116

        return cls.from_xyz(
            inv_f(lp + a / 500),
            inv_f(lp),
            inv_f(lp - b / 200),
        )



class HSVMixin:
    """Mixin to add HSV conversion functions."""

    def to_hsv(self) -> tuple[int | float]:
        """
        Converts the Colour object into HSV.

        Returns:
            tuple[int | float] consisting of:
                hue (0 - 360): in degrees
                saturation (0.0 - 1.0): as decimal percentage
                value (0.0 - 1.0): as decimal percentage
        """
        r, g, b = self.to_rgbd()

        M = max(r, g, b)
        m = min(r, g, b)
        d = M - m

        if d == 0:
            h = 0
        elif M == r:
            h = 60 * ((g - b) / d % 6)
        elif M == g:
            h = 60 * ((b - r) / d + 2)
        else:
            h = 60 * ((r - g) / d + 4)

        return (h, d / M, M)

    @classmethod
    def from_hsv(cls, h: int | float, s: int | float, v: int | float):
        """
        Creates a Colour object from a HSV colour.

        Args:
            h: int | float (0 - 360)
                Hue in degrees.
            s: int | float (0.0 - 1.0)
                Saturation as a decimal percentage.
            v: int | float (0.0 - 1.0)
                Value as a decimal percentage.

        Returns: Colour
        """
        c = v * s
        x = c * (1 - abs(h / 60 % 2 - 1))
        m = v - c

        if 0 <= h < 60:
            rt, gt, bt = (c, x, 0)
        elif 60 <= h < 120:
            rt, gt, bt = (x, c, 0)
        elif 120 <= h < 180:
            rt, gt, bt = (0, c, x)
        elif 180 <= h < 240:
            rt, gt, bt = (0, x, c)
        elif 240 <= h < 300:
            rt, gt, bt = (x, 0, c)
        else:
            rt, gt, bt = (c, 0, x)

        return cls(
            (rt + m) * 255,
            (gt + m) * 255,
            (bt + m) * 255,
        )



class HexMixin:
    """Mixin to add hex conversion functions."""

    def to_hex(self, with_hashtag: bool = True) -> str:
        """
        Converts the Colour object into a hex string.

        Args:
            with_hashtag: bool = True
                If true, returns hex string with a leading hashtag.

        Returns: string
        """
        hex_string = (
            f"{'#' if with_hashtag else ''}{self.r:02x}{self.g:02x}{self.b:02x}"
        )

        if self.a != 255:
            hex_string += f"{self.a:02x}"

        return hex_string

    @classmethod
    def from_hex(cls, hex_string: str):
        """
        Creates a Colour object from a hex string.

        Args:
            hex_string: string
                Hexadecimal colour string with or without leading hashtag.

        Returns: Colour
        """
        hex_string = hex_string.lstrip("#")
        if len(hex_string) == 3:
            hex_string = hex_string * 2

        return cls(
            int(hex_string[:2], 16),
            int(hex_string[2:4], 16),
            int(hex_string[4:6], 16),
            int(hex_string[6:8], 16) if len(hex_string) >= 8 else 255,
        )



class CMYKMixin:
    """Mixin to add CMYK conversion functions."""

    def to_cmyk(self) -> tuple[int | float]:
        """
        Converts the Colour object into a CMYK colour.

        Returns: tuple[int | float (0.0 - 1.0)] consisting of CMYK values as a decimal percentage.
        """
        rt, gt, bt = self.to_rgbd()
        k = 1 - max(rt, gt, bt)
        c = (1 - rt - k) / (1 - k)
        m = (1 - gt - k) / (1 - k)
        y = (1 - bt - k) / (1 - k)

        return (c, m, y, k)

    @classmethod
    def from_cmyk(cls, c: int | float, m: int | float, y: int | float, k: int | float):
        """
        Creates a Colour object from a CMYK colour.

        Args:
            c: int | float (0.0 - 1.0)
                Cyan channel as a decimal percentage.
            m: int | float (0.0 - 1.0)
                Magenta channel as a decimal percentage.
            y: int | float (0.0 - 1.0)
                Yellow channel as a decimal percentage.
            k: int | float (0.0 - 1.0)
                Black key channel as a decimal percentage.

        Returns: Colour
        """
        return cls(
            255 * (1 - c) * (1 - k),
            255 * (1 - m) * (1 - k),
            255 * (1 - y) * (1 - k),
        )



class RGBMixin:
    """Mixin to add RGB / RGBA conversion functions."""

    def to_rgb(self) -> tuple[int | float]:
        """
        Converts the Colour object into RGB.

        Returns:
            tuple[int | float (0 - 255)] consisting of RGB values.
        """
        return (self.r, self.g, self.b)

    def to_rgba(self) -> tuple[int | float]:
        """
        Converts the Colour object into RGBA.

        Returns:
            tuple[int | float (0 - 255)] consisting of RGBA values.
        """
        return (self.r, self.g, self.b, self.a)

    def to_rgbd(self) -> tuple[int | float]:
        """
        Converts the Colour object into decimal RGB.

        Returns:
            tuple[int | float (0.0 - 1.0)] consisting of decimal RGB values.
        """
        return (self.r / 255, self.g / 255, self.b / 255)

    def to_rgbad(self) -> tuple[int | float]:
        """
        Converts the Colour object into decimal RGBA.

        Returns:
            tuple[int | float (0.0 - 1.0)] consisting of decimal RGBA values.
        """
        return (self.r / 255, self.g / 255, self.b / 255, self.a / 255)

    @classmethod
    def from_rgb(
        cls, r: int | float, g: int | float, b: int | float, a: int | float = 255
    ):
        """
        Creates a Colour object from an 8-bit RGBA colour.

        Args:
            r: int | float (0 - 255)
                Red channel value.
            g: int | float (0 - 255)
                Green channel value.
            b: int | float (0 - 255)
                Blue channel value.
            a: int | float (0 - 255) = 255
                Alpha channel value.

        Returns: Colour
        """
        return cls(r, g, b, a)

    @classmethod
    def from_rgbd(
        cls, r: int | float, g: int | float, b: int | float, a: int | float = 1
    ):
        """
        Creates a Colour object from a decimal RGBA colour (0.0 - 1.0).

        Args:
            r: int | float (0.0 - 1.0)
                Normalized red channel value.
            g: int | float (0.0 - 1.0)
                Normalized green channel value.
            b: int | float (0.0 - 1.0)
                Normalized blue channel value.
            a: int | float (0.0 - 1.0) = 1
                Normalized alpha channel value.

        Returns: Colour
        """
        return cls(r * 255, g * 255, b * 255, a * 255)



class HSLMixin:
    """Mixin to add HSL conversion functions."""

    def to_hsl(self) -> tuple[int | float]:
        """
        Converts the Colour object into HSL.

        Returns:
            tuple[int | float] consisting of:
                hue (0 - 360): in degrees
                saturation (0.0 - 1.0): as decimal percentage
                luminance (0.0 - 1.0): as decimal percentage
        """
        r, g, b = self.to_rgbd()

        M = max(r, g, b)
        m = min(r, g, b)
        d = M - m

        if d == 0:
            h = 0
        elif M == r:
            h = 60 * ((g - b) / d % 6)
        elif M == g:
            h = 60 * ((b - r) / d + 2)
        else:
            h = 60 * ((r - g) / d + 4)

        l = (M + m) / 2
        s = 0 if d == 0 else d / (1 - abs(2 * l - 1))

        return (h, s, l)

    @classmethod
    def from_hsl(cls, h: int | float, s: int | float, l: int | float):
        """
        Creates a Colour object from a HSL colour.

        Args:
            h: int | float (0 - 360)
                Hue in degrees.
            s: int | float (0.0 - 1.0)
                Saturation as a decimal percentage.
            l: int | float (0.0 - 1.0)
                Luminance as a decimal percentage.

        Returns: Colour
        """
        d = s * (1 - abs(2 * l - 1))
        m = 255 * (l - d / 2)
        x = d * (1 - abs((h / 60) % 2 - 1))

        if 0 <= h < 60:
            r = 255 * d + m
            g = 255 * x + m
            b = m
        elif 60 <= h < 120:
            r = 255 * x + m
            g = 255 * d + m
            b = m
        elif 120 <= h < 180:
            r = m
            g = 255 * d + m
            b = 255 * x + m
        elif 180 <= h < 240:
            r = m
            g = 255 * x + m
            b = 255 * d + m
        elif 240 <= h < 300:
            r = 255 * x + m
            g = m
            b = 255 * d + m
        else:
            r = 255 * d + m
            g = m
            b = 255 * x + m

        return cls(r, g, b)
